/*
 * sqltoast
 *
 * Copyright (c) 2010 Jay Pipes <jaypipes@gmail.com>
 * All rights reserved.
 *
 * Use and distribution licensed under the BSD license. See the
 * COPYING file in the root project directory for full text.
 */

/**
 * @file Grammar for parsing CREATE SCHEMA statements
 */

#ifndef SQLTOAST_PARSER_CREATE_SCHEMA_H
#define SQLTOAST_PARSER_CREATE_SCHEMA_H

#include "sqltoast/visibility.h"
#include "sqltoast/common.h"
#include "sqltoast/parser/common_rules.h"
#include "sqltoast/parser/identifier.h"
#include "sqltoast/parser/specifier.h"
#include "sqltoast/ast.h"
#include "sqltoast/nodes/statements/create_schema.h"

#include <boost/config/warning_disable.hpp>
#include <boost/spirit/include/qi.hpp>
#include <boost/spirit/include/phoenix_core.hpp>
#include <boost/spirit/include/phoenix_operator.hpp>
#include <boost/fusion/include/adapt_struct.hpp>
#include <boost/fusion/include/unused.hpp>

#include <string>

namespace sqltoast
{
namespace parser
{
using boost::fusion::unused_type;

struct create_schema_parser_attr
{
  bool if_not_exists;  /* if not exists? */
  non_qualified_identifier_attr schema_identifier; /* schema identifier */
  boost::optional<charset_specifier_attr> charset_specifier;/* optional charset name */
  boost::optional<collation_specifier_attr> collation_specifier;/* optional collation specifier */
};

/**
 * Parser for CREATE SCHEMA statements
 *
 * @detail
 *
 * Class template for a parser with a synthesized
 * attribute of an ast-node of type statements::create_schema.
 *
 * A CREATE SCHEMA statement can take the following EBNF form:
 * 
 * create_schema : CREATE schema_or_database if_not_exists identifier charset collation
 *
 * schema_or_database : SCHEMA 
 *                    | DATABASE
 *
 * if_not_exists : IF NOT EXISTS
 *               | empty
 *
 * collation : COLLATE identifier
 *           | empty
 *
 * charset : DEFAULT CHARACTER SET identifier
 *         | CHARACTER SET identifier
 *         | empty
 */
template<class iterator_type, class space_type>
struct create_schema_parser: boost::spirit::qi::grammar<iterator_type,
                                                        create_schema_parser_attr(),
                                                        space_type >
{
  create_schema_parser() : create_schema_parser::base_type(start)
  {
    using boost::spirit::qi::char_;
    using boost::spirit::qi::alpha;
    using boost::spirit::qi::alnum;
    using boost::spirit::ascii::no_case;
    using boost::spirit::eps;
    using boost::spirit::eoi;
    using boost::spirit::omit;
    using boost::spirit::qi::_val;
    using boost::spirit::lexeme;

    start = omit[no_case["create"]]
            >> schema_or_database 
            >> if_not_exists 
            >> non_qualified_identifier_rule
            >> -charset_specifier_rule
            >> -collation_specifier_rule 
            >> eoi;
    schema_or_database = omit[no_case["schema"] | no_case["database"]];
  }

  boost::spirit::qi::rule<iterator_type, create_schema_parser_attr(), space_type > start;
  boost::spirit::qi::rule<iterator_type, unused_type, space_type > schema_or_database;
};

/* Our create schema parsers, created during library initialization time */
create_schema_parser<std::string::iterator, boost::spirit::qi::ascii::space_type> basic_ascii_create_schema_parser;

/**
 * Parses a string for the CREATE SCHEMA statement
 *
 * @param[in] Start of string to parse
 * @param[in] End of string to parse
 * @param[out] Reference to output attribute
 */
template<class iterator_type>
SQLTOAST_API
result_code parse_create_schema(iterator_type start,
                                iterator_type end,
                                create_schema_parser_attr& output)
{
  if (boost::spirit::qi::phrase_parse(start,
                                      end,
                                      basic_ascii_create_schema_parser,
                                      boost::spirit::qi::ascii::space,
                                      output))
    return SUCCESS;
  return SYNTAX_ERROR;
}

/**
 * Parses a string for the CREATE SCHEMA statement
 *
 * @param[in] Start of string to parse
 * @param[in] End of string to parse
 * @param[out] Reference to output AST
 */
template<class iterator_type>
SQLTOAST_API
result_code parse_create_schema(iterator_type start,
                                iterator_type end,
                                ast& output)
{
  create_schema_parser_attr attr;

  if (boost::spirit::qi::phrase_parse(start,
                                      end,
                                      basic_ascii_create_schema_parser,
                                      boost::spirit::qi::ascii::space,
                                      attr))
  {
    boost::shared_ptr<nodes::identifier> ptr_identifier = 
      boost::shared_ptr<nodes::identifier>(
        new nodes::identifier(attr.schema_identifier));
    boost::shared_ptr<nodes::statement> ptr_statement = 
      boost::shared_ptr<nodes::statement>(
        new nodes::statements::create_schema(ptr_identifier,
                                             attr.if_not_exists,
                                             attr.charset_specifier,
                                             attr.collation_specifier));
    output.set_root(ptr_statement);
    return SUCCESS;
  }
  return SYNTAX_ERROR;
}

} // namespace parser
} // namespace sqltoast

BOOST_FUSION_ADAPT_STRUCT(
  sqltoast::parser::create_schema_parser_attr,
  (bool, if_not_exists)
  (sqltoast::parser::non_qualified_identifier_attr, schema_identifier)
  (boost::optional<sqltoast::parser::charset_specifier_attr>, charset_specifier)
  (boost::optional<sqltoast::parser::collation_specifier_attr>, collation_specifier)
)

#endif /* SQLTOAST_PARSER_CREATE_SCHEMA_H */
